/*******************************************************************************
 * Copyright (c) 2003, 2007 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/

/*
 * @(#) $Id: PipeListenerThread.java 2966 2008-12-12 14:32:23Z gub $
 *
 *
 *
 *
 *
 */
package at.spardat.xma.session;

import java.io.IOException;
import java.util.Enumeration;
import java.util.Properties;
import java.util.StringTokenizer;

import at.spardat.xma.boot.logger.LogLevel;
import at.spardat.xma.boot.logger.Logger;
import at.spardat.xma.pipes.XMAPipe;

/**
 * @author s3460
 * @since version_number
 *
 * Thread creating a named pipe and listening for a client. If a client
 * connects, the byte stream from the client is read. Then an ExternalEvent is
 * instantiated and XMAClienSession.notifyExternalEvent() is called.
 * ExternalEvent return values are written back to the client as byte[] stream.
 *
 */
public class PipeListenerThread extends Thread {

    //private static final String PIPE_NAME = "\\\\.\\pipe\\mynamedpipe";

    private static final String DELIM = "\b";

    private static final String XMA_TIMEOUT = "at.spardat.xma.timeout";

    private XMAPipe pipe;

    private boolean stopThread;

    private Logger log;

    private XMASessionClient session;

    /**
     *
     */
    public PipeListenerThread(XMASessionClient session, XMAPipe pipe) {
        super();
        this.session = session;
        this.pipe = pipe;
        setDaemon(true);
        log = Logger.getLogger("Pipe-Listener: " + pipe.getPipeName());
    }

    /*
     * (non-Javadoc)
     *
     * @see java.lang.Runnable#run()
     */
    public void run() {
        while (!stopThread) {
            try {
                pipe.open();
                byte[] in = pipe.read();
                byte[] out = notify(in);

                if (out != null && out.length > 0 ) {
                    pipe.write(out);
                }
            } catch (IOException e) {
                log.log(LogLevel.WARNING, "Error pipe communication :", e);
            } finally {
                try {
                    pipe.close();
                } catch (IOException e1) {
                    log.log(LogLevel.WARNING, "Error closing pipe :", e1);
                }
            }
        }

        try {
            pipe.destroyPipe();
        } catch (IOException e) {
            log.log(LogLevel.SEVERE, "can't destroy XMAPipe:", e);
        }
    }

    /**
     * Notifies the ExternalEvent listener. Calls
     * XMAClienSession.notifyExternalEvent().
     *
     * @param in
     * @return @since version_number
     * @author s3460
     */
    private byte[] notify(byte[] in) {
        final ExternalEvent event = new ExternalEvent(byteToProperties(in));

        long timeout = Long.parseLong(event.getProperty(XMA_TIMEOUT,"0"));
        event.getProperties().remove(XMA_TIMEOUT);

        class Runner implements Runnable {
            Properties prop;

            public void run() {
                prop = session.notifyExternalEvent(event);
                synchronized (PipeListenerThread.this){//notifies outer class to end wait().
                    PipeListenerThread.this.notify();  //must be called in a block synchronized on the outer class
                }
            }
        }

        Runner runner = new Runner();
        //session.display.syncExec(runner);
        session.display.asyncExec(runner);
        try {
            synchronized (this){ //waits for timeout or notify() on this class (= outer class)
            wait(timeout);       //must be called in a block synchronized on this class
            }
        } catch (InterruptedException e) {
            log.log(LogLevel.INFO, "Exception by wait(timeout) on ExternalEvent Impl", e);
        }
        if (runner.prop != null) {
            return propertiesToByte(runner.prop);
        } else {
            return null;
        }

    }

    /**
     * @param prop
     * @return @since version_number
     * @author s3460
     */
    private byte[] propertiesToByte(Properties prop) {
        StringBuffer buffer = new StringBuffer();
        for (Enumeration iter = prop.keys(); iter.hasMoreElements();) {
            String key = (String) iter.nextElement();
            String value = prop.getProperty(key);
            buffer.append(key);
            buffer.append(DELIM);
            buffer.append(value);
            buffer.append(DELIM);
        }

        //Use the default encoding as the native parts (xmalauncher/xmaserver)
        //cannot deal with the UTF-8 encoding
        return buffer.toString().getBytes();
    }

    /**
     * @param in
     * @return @since version_number
     * @author s3460
     */
    private Properties byteToProperties(byte[] in) {
        boolean inDelim = false;
        boolean isKey = true;
        Properties prop = new Properties();
        String key = null;
        String string = new String(in, 0, in.length);
        StringTokenizer st = new StringTokenizer(string,DELIM,true);
        while (st.hasMoreTokens()) {
            if(inDelim) {
                st.nextToken();
                inDelim=false;
            } else {
                if (isKey) {
                    key = st.nextToken();
                    if(DELIM.equals(key)) {
                        key="";
                    } else {
                        inDelim=true;
                    }
                } else {
                    String value = st.nextToken();
                    if(DELIM.equals(value)) {
                        value="";
                    } else {
                        inDelim=true;
                    }
                    prop.setProperty(key, value);
                }
                isKey = !isKey;
            }
        }

        return prop;
    }

    /**
     * Stops the PipeListenerThread (the Loop in run()).
     */
    public void stopThread() {
        this.stopThread = true;
    }
}
